@@ -5,11 +5,11 @@
 
 use crate::raw::{
     config::Config,
-    error::{Error, Fallible},
+    error::{Error, Fallible, Pollable},
     security,
 };
 use core::{
-    convert::TryInto,
+    convert::{TryFrom, TryInto},
     fmt,
     ptr::NonNull,
     task::{Poll, Waker},
@@ -18,7 +18,73 @@ use libc::c_void;
 use s2n_tls_sys::*;
 use std::{ffi::CStr, mem};
 
-pub use s2n_tls_sys::s2n_mode;
+use s2n_tls_sys::s2n_mode;
+
+macro_rules! static_const_str {
+    ($c_chars:expr) => {
+        unsafe { CStr::from_ptr($c_chars) }
+            .to_str()
+            .map_err(|_| Error::InvalidInput)
+    };
+}
+
+#[derive(Debug, PartialEq)]
+pub enum CallbackResult {
+    Success,
+    Failure,
+}
+
+impl From<CallbackResult> for s2n_status_code::Type {
+    fn from(input: CallbackResult) -> s2n_status_code::Type {
+        match input {
+            CallbackResult::Success => s2n_status_code::SUCCESS,
+            CallbackResult::Failure => s2n_status_code::FAILURE,
+        }
+    }
+}
+
+#[derive(Debug, PartialEq)]
+pub enum Mode {
+    Server,
+    Client,
+}
+
+impl From<Mode> for s2n_mode::Type {
+    fn from(input: Mode) -> s2n_mode::Type {
+        match input {
+            Mode::Server => s2n_mode::SERVER,
+            Mode::Client => s2n_mode::CLIENT,
+        }
+    }
+}
+
+#[non_exhaustive]
+#[derive(Debug, PartialEq)]
+pub enum Version {
+    SSLV2,
+    SSLV3,
+    TLS10,
+    TLS11,
+    TLS12,
+    TLS13,
+}
+
+impl TryFrom<s2n_tls_version::Type> for Version {
+    type Error = Error;
+
+    fn try_from(input: s2n_tls_version::Type) -> Result<Self, Self::Error> {
+        let version = match input {
+            s2n_tls_version::SSLV2 => Self::SSLV2,
+            s2n_tls_version::SSLV3 => Self::SSLV3,
+            s2n_tls_version::TLS10 => Self::TLS10,
+            s2n_tls_version::TLS11 => Self::TLS11,
+            s2n_tls_version::TLS12 => Self::TLS12,
+            s2n_tls_version::TLS13 => Self::TLS13,
+            _ => return Err(Error::InvalidInput),
+        };
+        Ok(version)
+    }
+}
 
 pub struct Connection {
     connection: NonNull<s2n_connection>,
@@ -26,9 +92,17 @@ pub struct Connection {
 
 impl fmt::Debug for Connection {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        f.debug_struct("Connection")
-            // TODO add paths
-            .finish()
+        let mut debug = f.debug_struct("Connection");
+        if let Ok(handshake) = self.handshake_type() {
+            debug.field("handshake_type", &handshake);
+        }
+        if let Ok(cipher) = self.cipher_suite() {
+            debug.field("cipher_suite", &cipher);
+        }
+        if let Ok(version) = self.actual_protocol_version() {
+            debug.field("actual_protocol_version", &version);
+        }
+        debug.finish_non_exhaustive()
     }
 }
 
@@ -38,9 +112,10 @@ impl fmt::Debug for Connection {
 unsafe impl Send for Connection {}
 
 impl Connection {
-    pub fn new(mode: s2n_mode::Type) -> Self {
+    pub fn new(mode: Mode) -> Self {
         crate::raw::init::init();
-        let connection = unsafe { s2n_connection_new(mode).into_result() }.unwrap();
+
+        let connection = unsafe { s2n_connection_new(mode.into()).into_result() }.unwrap();
 
         unsafe {
             debug_assert! {
@@ -62,11 +137,11 @@ impl Connection {
     }
 
     pub fn new_client() -> Self {
-        Self::new(s2n_mode::CLIENT)
+        Self::new(Mode::Client)
     }
 
     pub fn new_server() -> Self {
-        Self::new(s2n_mode::SERVER)
+        Self::new(Mode::Server)
     }
 
     /// # Safety
@@ -287,10 +362,52 @@ impl Connection {
     pub fn negotiate(&mut self) -> Poll<Result<&mut Self, Error>> {
         let mut blocked = s2n_blocked_status::NOT_BLOCKED;
 
-        match unsafe { s2n_negotiate(self.connection.as_ptr(), &mut blocked).into_result() } {
-            Ok(_) => Ok(self).into(),
-            Err(err) if err.kind() == s2n_error_type::BLOCKED => Poll::Pending,
-            Err(err) => Err(err).into(),
+        unsafe {
+            s2n_negotiate(self.connection.as_ptr(), &mut blocked)
+                .into_poll()
+                .map_ok(|_| self)
+        }
+    }
+
+    /// Encrypts and sends data on a connection where
+    /// [negotiate](`Self::negotiate`) has succeeded.
+    ///
+    /// Returns the number of bytes written, and may indicate a partial write.
+    pub fn send(&mut self, buf: &[u8]) -> Poll<Result<usize, Error>> {
+        let mut blocked = s2n_blocked_status::NOT_BLOCKED;
+        let buf_len: isize = buf.len().try_into().map_err(|_| Error::InvalidInput)?;
+        let buf_ptr = buf.as_ptr() as *const ::libc::c_void;
+        unsafe { s2n_send(self.connection.as_ptr(), buf_ptr, buf_len, &mut blocked).into_poll() }
+    }
+
+    /// Reads and decrypts data from a connection where
+    /// [negotiate](`Self::negotiate`) has succeeded.
+    ///
+    /// Returns the number of bytes read, and may indicate a partial read.
+    /// 0 bytes returned indicates EOF due to connection closure.
+    pub fn recv(&mut self, buf: &mut [u8]) -> Poll<Result<usize, Error>> {
+        let mut blocked = s2n_blocked_status::NOT_BLOCKED;
+        let buf_len: isize = buf.len().try_into().map_err(|_| Error::InvalidInput)?;
+        let buf_ptr = buf.as_ptr() as *mut ::libc::c_void;
+        unsafe { s2n_recv(self.connection.as_ptr(), buf_ptr, buf_len, &mut blocked).into_poll() }
+    }
+
+    /// Attempts to flush any data previously buffered by a call to [send](`Self::negotiate`).
+    pub fn flush(&mut self) -> Poll<Result<&mut Self, Error>> {
+        self.send(&[0; 0]).map_ok(|_| self)
+    }
+
+    /// Attempts a graceful shutdown of the TLS connection.
+    ///
+    /// The shutdown is not complete until the necessary shutdown messages
+    /// have been successfully sent and received. If the peer does not respond
+    /// correctly, the graceful shutdown may fail.
+    pub fn shutdown(&mut self) -> Poll<Result<&mut Self, Error>> {
+        let mut blocked = s2n_blocked_status::NOT_BLOCKED;
+        unsafe {
+            s2n_shutdown(self.connection.as_ptr(), &mut blocked)
+                .into_poll()
+                .map_ok(|_| self)
         }
     }
 
@@ -389,6 +506,29 @@ impl Connection {
         };
         Ok(())
     }
+
+    pub fn actual_protocol_version(&self) -> Result<Version, Error> {
+        let version = unsafe {
+            s2n_connection_get_actual_protocol_version(self.connection.as_ptr()).into_result()?
+        };
+        version.try_into()
+    }
+
+    pub fn handshake_type(&self) -> Result<&str, Error> {
+        let handshake = unsafe {
+            s2n_connection_get_handshake_type_name(self.connection.as_ptr()).into_result()?
+        };
+        // The strings returned by s2n_connection_get_handshake_type_name
+        // are static and immutable after they are first calculated
+        static_const_str!(handshake)
+    }
+
+    pub fn cipher_suite(&self) -> Result<&str, Error> {
+        let cipher = unsafe { s2n_connection_get_cipher(self.connection.as_ptr()).into_result()? };
+        // The strings returned by s2n_connection_get_cipher
+        // are static and immutable since they are const fields on static const structs
+        static_const_str!(cipher)
+    }
 }
 
 #[derive(Default)]
